#ifndef BOARD_H
#define BOARD_H

#include "common.h"
#include "Package.h"
#include "Robot.h"
#include "Path.h"

class Board
{
public:
  enum PosType { eOpen=0, eHome, eWater, eWall };

  Board(vector<string> data)
  {
    int xlen = data[0].size();
    mData.resize(xlen+2);
    int ylen = data.size();
    int i;
    for (i = 0; i < mData.size(); i++)
      mData[i].resize(ylen+2);

    for (i = 0; i < mData.size(); i++)
    {
      mData[i][0] = eWall;
      mData[i][mData[i].size()-1] = eWall;
    }
    for (i = 0; i < mData[0].size(); i++)
    {
      mData[0][i] = eWall;
      mData[mData.size()-1][i] = eWall;
    }

    for (i = 0; i < data.size(); i++)
      for (int j = 0; j < data[i].size(); j++)
      {
        switch(data[i][j])
        {
          case '.': mData[j+1][i+1] = eOpen; break;
          case '@': mData[j+1][i+1] = eHome; mHomes.push_back(Point(j+1, i+1)); break;
          case '~': mData[j+1][i+1] = eWater; break;
          case '#': mData[j+1][i+1] = eWall; break;
        }
      }

    buildTree();

    //printTree(mNodes[0]);
  }

class Node
{
public:
  Node() : mVisit(false) {}
  Node(const Point& ul, const Point& lr) : mUL(ul), mLR(lr), mVisit(false) {}

  bool hasPoint(const Point& pt)
  { return pt.x() >= mUL.x() && pt.y() >= mUL.y() &&
           pt.x() <= mLR.x() && pt.y() <= mLR.y(); }

  bool visited() const   { return mVisit; }
  void markVisit(bool b) { mVisit = b; }

  Point ul() const { return mUL; }
  Point lr() const { return mLR; }
  Point center() const { return Point((mUL.x() + mLR.x()) / 2,
                                      (mUL.y() + mLR.y()) / 2); }

  void addX(int i) { mLR = Point(mLR.x() + i, mLR.y()); }
  void addY(int i) { mLR = Point(mLR.x(), mLR.y() + i); }

  const vector<Node*>& adjs() const { return mAdj; }

  void addAdj(Node* n)
  {
    if (find(mAdj.begin(), mAdj.end(), n) != mAdj.end())
      return;

    if (
        hasPoint(Point(n->mLR.x() + 1, n->mLR.y()    )) || // Check to the left
        hasPoint(Point(n->mLR.x(),     n->mLR.y() + 1)) || // Check to the top
        hasPoint(Point(n->mUL.x() - 1, n->mUL.y()    )) || // Check to the right
        hasPoint(Point(n->mUL.x(),     n->mUL.y() - 1)) || // Check to the bottom
        n->hasPoint(Point(mLR.x() + 1, mLR.y()    )) || // Check to the left
        n->hasPoint(Point(mLR.x(),     mLR.y() + 1)) || // Check to the top
        n->hasPoint(Point(mUL.x() - 1, mUL.y()    )) || // Check to the right
        n->hasPoint(Point(mUL.x(),     mUL.y() - 1))    // Check to the bottom
       )
      mAdj.push_back(n);
  }

protected:
  Point mUL;
  Point mLR;
  bool mVisit;
  vector<Node*> mAdj;
};

  void buildTree()
  {
    // Create all nodes
    for (int j = 0; j < mData[0].size(); j++)
      for (int i = 0; i < mData.size(); i++)
        makeNode(Point(i, j));

    // Connect them
    for (vector<Node*>::iterator it1 = mNodes.begin();
         it1 != mNodes.end();
         it1++)
      for (vector<Node*>::iterator it2 = mNodes.begin();
           it2 != mNodes.end();
           it2++)
        (*it1)->addAdj(*it2);
  }

  void printTree(Node* n)
  {
    if (n->visited())
      return;

    n->markVisit(true);
    cout << "Node: " << n->ul() << "-" << n->lr() << endl << "  Adj:";
    for (int i = 0; i < n->adjs().size(); i++)
      cout << n->adjs()[i]->ul() << "-" << n->adjs()[i]->lr() << ",";

    cout << endl;

    for (int j = 0; j < n->adjs().size(); j++)
      printTree(n->adjs()[j]);
  }

  Board::Node* findNode(const Point& pt)
  {
    if (!canPass(pt))
      return NULL;

    for (vector<Node*>::iterator it = mNodes.begin();
         it != mNodes.end();
         ++it)
      if ((*it)->hasPoint(pt))
        return *it;

    return NULL;
  }

  Board::Node* makeNode(const Point& pt)
  {
    if (!canPass(pt))
      return NULL;

    Node* n = findNode(pt);
    if (n)
      return n;

    Node* ret = new Node(pt, pt);
    while (canPass(ret->lr()) && !findNode(ret->lr()))
      ret->addX(1);
    ret->addX(-1);

    bool rowValid = true;
    while (rowValid)
    {
      for (int i = ret->ul().x(); i <= ret->lr().x(); i++)
      {
        Point p(i, ret->lr().y());
        if (!canPass(p) || findNode(p))
          rowValid = false;
      }

      if (rowValid)
        ret->addY(1);
    }
    ret->addY(-1);

    mNodes.push_back(ret);
    return ret;
  }

  void clearVisited()
  {
    for (vector<Node*>::iterator it = mNodes.begin();
         it != mNodes.end();
         ++it)
      (*it)->markVisit(false);
  }

  vector<Node*> getNodePath(Node* n, const Point& ptb)
  {
    if (n->hasPoint(ptb))
    {
      vector<Node*> ret;
      ret.push_back(n);
      return ret;
    }

    if (n->visited())
      return vector<Node*>();

    n->markVisit(true);
    vector<Node*> ret;
    for (vector<Node*>::const_iterator it = n->adjs().begin();
         it != n->adjs().end();
         it++)
    {
      vector<Node*> tmp(getNodePath(*it, ptb));
      if (tmp.size() && tmp.size() < ret.size() ||
          tmp.size() && !ret.size())
        ret = tmp;
    }

    if (ret.size())
    {
      ret.push_back(n);
    }

    return ret;
  }

  Path getPathTree(const Point& pta, const Point& ptb)
  {
    clearVisited();
    Node* st = findNode(pta);
    vector<Node*> npath(getNodePath(st, ptb));

    if (npath.size() <= 1)
      return getShortestPathSimple(pta, ptb);

    Path ret(getShortestPathSimple(npath[0]->center(), ptb));
    if (!ret.size())
      cout << "TREE ERROR: No Path from " << ptb << " to "
                                          << npath[0]->center() << endl;

    for (int i = 1; i < npath.size(); i++)
    {
      Path tmp(getShortestPathSimple(npath[i]->center(), npath[i-1]->center()));
      if (!tmp.size())
        cout << "TREE ERROR: No Path from " << npath[i-1]->center() << " to "
                                            << npath[i]->center() << endl;
      ret.insert(ret.end(), tmp.begin(), tmp.end());
    }

    Path tmp(getShortestPathSimple(pta, npath.back()->center()));
    if (!tmp.size())
      cout << "TREE ERROR: No Path from " << npath.back()->center() << " to "
                                          << pta << endl;
    ret.insert(ret.end(), tmp.begin(), tmp.end());

    if (ret.size())
    {
      for (Path::iterator it = ret.begin();
           (it + 1) != ret.end(); )
      {
        if (*it == *(it + 1))
        {
          it = ret.erase(it);
        }
        else
          it++;
      }
    }

    return ret;
  }

  void moveNorth(int id)
  {
    move(id, getRobot(id)->pos() + Point(0, 1));
  }

  void moveSouth(int id)
  {
    move(id, getRobot(id)->pos() - Point(0, 1));
  }

  void moveEast(int id)
  {
    move(id, getRobot(id)->pos() + Point(1, 0));
  }

  void moveWest(int id)
  {
    move(id, getRobot(id)->pos() - Point(1, 0));
  }

  void move(int id, const Point& dest)
  {
    getRobot(id)->pos(dest);
  }

  Robot* getRobot(int id)
  {
    for (vector<Robot*>::iterator it = mRobots.begin();
         it != mRobots.end();
         it++)
    {
      if ((*it)->id() == id)
        return *it;
    }

    Robot* ret = new Robot(id);
    mRobots.push_back(ret);
    return ret;
  }

  Package getPackage(int id, bool bRemove = false)
  {
    for (map<Point, vector<Package> >::iterator it = mPkgs.begin();
         it != mPkgs.end();
         it++)
    {
      for (vector<Package>::iterator it1 = it->second.begin();
           it1 != it->second.end();
           it1++)
      {
        if (it1->id() == id)
        {
          Package ret(*it1);
          if (bRemove)
            it->second.erase(it1);
          return ret;
        }
      }
    }

    for (vector<Robot*>::iterator it1 = mRobots.begin();
         it1 != mRobots.end();
         it1++)
    {
      Package p((*it1)->getPkg(id));
      if (p.id() == id)
        return p;
    }

    return Package(id);
  }

  void addPackage(const Point& pt, const Package& p)
  { mPkgs[pt].push_back(p); }

  vector<Package> getPackagesAt(const Point& pt)
  { return mPkgs[pt]; }

  void clearPackagesAt(const Point& pt)
  { mPkgs[pt].clear(); }

  void addRobot(Robot* r)
  {
    mRobots.push_back(r);
  }

  void verifyRobots(vector<int> ids)
  {
    for (vector<Robot*>::iterator it = mRobots.begin();
         it != mRobots.end();
         )
    {
      if (find(ids.begin(), ids.end(), (*it)->id()) == ids.end())
        it = mRobots.erase(it);
      else
        it++;
    }
  }

  void pickup(int id, int pkgid)
  {
    getRobot(id)->pickup(getPackage(pkgid, true));
  }

  void drop(int id, int pkgid)
  {
    getRobot(id)->drop(pkgid);
  }

  bool canPass(const Point& pt)
  {
    if (pt.x() >= mData.size() || pt.y() >= mData[pt.x()].size())
    {
      cout << "Can't pass: " << pt << endl;
      return false;
    }

    PosType p = mData[pt.x()][pt.y()];
    return p == eOpen || p == eHome;
  }

  static char getDir(const Point& pt1, const Point& pt2)
  {
    if (pt2.x() < pt1.x())
      return 'W';
    if (pt2.x() > pt1.x())
      return 'E';
    if (pt2.y() < pt1.y())
      return 'S';
    if (pt2.y() > pt1.y())
      return 'N';

    cout << endl << "ERROR: (" << pt1.x() << "," << pt1.y() << ") == ("
         << pt2.x() << "," << pt2.y() << ")" << endl;

    return ' ';
  }

  map<pair<Point, Point>, Path> past;

  // Returns reversed path
  Path getShortestPathSimple(const Point& pta, 
                             const Point& ptb, 
                             double dist = 1e6)
  {
    if (pta == ptb)
      return Path(pta);

    if (!canPass(pta))
      return Path();

#if 0
    map<pair<Point, Point>, Path>::iterator it = past.find(make_pair(pta, ptb));
    if (it != past.end())
      return it->second;
    it = past.find(make_pair(ptb, pta));
    if (it != past.end())
    {
      Path ret(it->second);
      reverse(ret.begin(), ret.end());
      return ret;
    }
#endif

    double d = sqrt((pta.x()-ptb.x())*(pta.x()-ptb.x()) + 
                    (pta.y()-ptb.y())*(pta.y()-ptb.y()));
    if (d >= dist)
      return Path();

    Path p1 = getShortestPathSimple(pta + Point(0, 1), ptb, d);
    Path p2 = getShortestPathSimple(pta - Point(0, 1), ptb, d);
    Path p3 = getShortestPathSimple(pta + Point(1, 0), ptb, d);
    Path p4 = getShortestPathSimple(pta - Point(1, 0), ptb, d);
    Path& ret = p1;
    if (p2.size() && p2.size() < ret.size() ||
        p2.size() && !ret.size())
      ret = p2;
    if (p3.size() && p3.size() < ret.size() ||
        p3.size() && !ret.size())
      ret = p3;
    if (p4.size() && p4.size() < ret.size() ||
        p4.size() && !ret.size())
      ret = p4;

    if (ret.size())
    {
      ret.push_back(pta);
      past[make_pair(pta, ptb)] = ret;
    }

    return ret;
  }

  map<pair<Point, Point>, Path> pasta;

  Path getSomePath(const Point& pta, 
                   const Point& ptb, 
                   vector<Point>& visited)
  {
    if (pta == ptb)
      return Path(pta);

    if (!canPass(pta))
    {
      return Path();
    }

#if 1
    map<pair<Point, Point>, Path>::iterator it = pasta.find(make_pair(pta, ptb));
    if (it != pasta.end())
      return it->second;
    it = pasta.find(make_pair(ptb, pta));
    if (it != pasta.end())
    {
      Path ret(it->second);
      reverse(ret.begin(), ret.end());
      return ret;
    }
#endif

    //cout << "(" << pta.x() << "," << pta.y() << ")";
    if (find(visited.begin(), visited.end(), pta) != visited.end())
      return Path();

    visited.push_back(pta);

    //cout << "n";

    Path p1 = getSomePath(pta + Point(0, 1), ptb, visited);
    Path p2 = getSomePath(pta - Point(0, 1), ptb, visited);
    Path p3 = getSomePath(pta + Point(1, 0), ptb, visited);
    Path p4 = getSomePath(pta - Point(1, 0), ptb, visited);
    Path& ret = p1;
    if (p2.size() && p2.size() < ret.size() ||
        p2.size() && !ret.size())
      ret = p2;
    if (p3.size() && p3.size() < ret.size() ||
        p3.size() && !ret.size())
      ret = p3;
    if (p4.size() && p4.size() < ret.size() ||
        p4.size() && !ret.size())
      ret = p4;

    if (ret.size())
    {
      ret.push_back(pta);
      pasta[make_pair(pta, ptb)] = ret;
    }

    return ret;
  }

  Path getShortestPathComplex(const Point& pta, 
                              const Point& ptb,
                              set<Point>& tried)
  {
    if (!canPass(pta) || !canPass(ptb))
      return Path();
    Path ret(getShortestPathSimple(pta, ptb));
    if (ret.size())
      return ret;

    vector<Point> visit;
    return getSomePath(pta, ptb, visit);
  }

  Path getShortestPath(const Point& pta, const Point& ptb)
  {
    set<Point> tried;
    Path ret(getPathTree(pta, ptb));
    reverse(ret.begin(), ret.end());
    if (ret.size())
    {
      //ret.print();
    }
    else
      cout << "No Path from (" << pta.x() << "," << pta.y() << ") to (" << ptb.x() <<
              "," << ptb.y() << ")" << endl;
    return ret;
#if 0
    if (pta == ptb)
      return Path(pta);

    if (past.find(make_pair(pta, ptb)) != past.end())
      return past[make_pair(pta, ptb)];

    Path tmp(path);
#endif
  }

  Path getPathToClosestPackage(const Point& st)
  {
    Path ret;
    for (map<Point, vector<Package> >::iterator it = mPkgs.begin();
         it != mPkgs.end();
         it++)
    {
      if (!it->second.size())
        continue;
      Path tst(getShortestPath(st, it->first));
      if (tst.size() && tst.size() < ret.size() ||
          tst.size() && !ret.size())
      {
        ret = tst;
      }
    }

    if (ret.size())
      return ret;

    return getPathToClosestHome(st);
  }

  void removeHome(const Point& pt)
  {
    for (vector<Point>::iterator it = mHomes.begin();
         it != mHomes.end();
         it++)
    {
      if (pt == *it)
      {
        mHomes.erase(it);
        return;
      }
    }
  }

  Path getPathToClosestHome(const Point& pt)
  {
    // Must have at least one home
    Path ret(getShortestPath(pt, mHomes[0]));
    for (int i = 1; i < mHomes.size(); i++)
    {
      Path tst(getShortestPath(pt, mHomes[i]));
      if (tst.size() && tst.size() < ret.size() ||
          tst.size() && !ret.size())
        ret = tst;
    }

    return ret;
  }

  void print()
  {
    cout << " ";
    for (int k = 0; k < mData.size(); k++)
      cout << k % 10;
    cout << endl;
    for (int i = 0; i < mData[0].size(); i++)
    {
      cout << i % 10;
      for (int j = 0; j < mData.size(); j++)
      {
        switch (mData[j][i])
        {
          case eOpen: cout << "."; break;
          case eHome: cout << "@"; break;
          case eWall: cout << "#"; break;
          case eWater: cout << "~"; break;
        }
      }
      cout << endl;
    }
  }

private:
  vector<vector<PosType> > mData;
  map<Point, vector<Package> > mPkgs;
  vector<Robot*>  mRobots; // Map?
  vector<Point>   mHomes;  // Cache, for speed

  vector<Node*>   mNodes;
};

#endif

